(* animals.ml: demonstrate a small class hierarchy of from the animals
   class. The main loop demonstrates how method calls dispatch to the
   specific function which is relevant to each animal. *)

open Printf;;

class virtual animal =                   (* virtual: some methods un-implementd *)
object(this)                             (* refer to object via 'this' *)
  method virtual id  : unit -> string    (* method not implemented *)
  method say () =                        (* implmented method *)
    printf "I'm a %s\n" (this#id ())
end;;

class fish =                             (* another class *)
object(me)                               (* refer to object via "me"  *)
inherit animal                           (* subclass of animal *)
  method id () = "fish"                  (* id method specified *)
end;;                                    (* say method inherited *)

class duck = object                      (* another class *)
  inherit animal                         (* subclass of animal *)
  method id  () = "duck"                 (* override both methods *)
  method say () =
    printf "quack\n"
end;;

class mascot = object                    (* subclass of duck *)
  inherit duck                           (* inherits id method *)
  method say () =                        (* overrides say method *)
    printf "Aflack!\n"
end;;

class fox =                              (* another class *)
object(self) inherit animal              (* refer to object via self, subclass of animal *)
  method id  () = "fox"                  (* override id method *)
  method say () =                        (* and say method *)
    printf "\nRing-ding-ding-ding-dingeringeding!\n";
    printf "Gering-ding-ding-ding-dingeringeding!\n";
    printf "Gering-ding-ding-ding-dingeringeding!\n"
end;;

let _ =                                  (* main function *)
  let animals = [|                       (* array of animals *)
      ((new fish)   :> animal);          (* "upcast" required to satisfy  *)
      ((new mascot) :> animal);          (* type checker: all array elems *)
      ((new duck)   :> animal);          (* elements of list are thus same *)
      ((new fox)    :> animal);          (* type through inheritance *)
    |]
  in
  let len = Array.length animals in
  for i=0 to len-1 do                    (* iterate over animals  *)
    let a = animals.(i) in
    printf "The %s says: " (a#id ());    (* methods dispatch to specific type *)
    a#say ();                            (* of animal printing different results *)
  done;
;;
